* SOUND stereo sample replay device driver for Qdos
* Classic and SMSQ/E, by Simon N Goodwin with thanks
* to Luca Pivato, Mark J Swift and Jonathan R Hudson
* Version 2e, slightly enhanced by Marcel Kilgus.
* Freely distributable only with source

sss_type   equ	    24		    One channel variable

* Check the SSS exists; if it does, remember where it lives

start	   move.l   #$70,a0	    Level 4 interrupt autovector offset

* Routine to determine CPU type - distinguishes 68020+ from earlier
* models. See 68XXX Programmer's Reference Manual, page 4-128

	   dc.l     $48e70001
;	    movem.l  a7,-(a7)	     Don't let this line get 'optimised'
	   cmpa.l   (a7)+,a7	    Was stored value pre-decremented?
	   beq.s    CPU16	    Zero if this is a 68008 or 68000

* If this comes out zero, this must be an old non-pipelined 68K.
* VBR will be assumed to be zero and not read.	This trick makes
* the code reliable on 68008 (QL) 68000 (stock 16 bit Amigas/STs)
* and 68020 or above. If you try to run it on a 68010 with system
* vectors relocated, it will fail. I can't quite please everyone.

	   trap     #0		    Supervisor mode
	   dc.l     $4e7ab801	    Magic instruction (JRH) movec vbr,a3
	   add.l    a3,a0	    Adjust for arbitrary vector base
	   andi.w   #$d8ff,sr	    Return to user mode

CPU16	   move.l   (a0),a0	    Get the level 5 autovector
	   subq.l   #8,a0	    Look back over two long words
	   cmpi.l   #'SSSS',(a0)    Check for magic marker before that
	   bne	    not_yet	    Sampled sound system not available

	   lea.l    sss_ptr(pc),a2
	   move.l   a0,(a2)	    Got SSS; remember where we found it

	   lea.l    Routines(pc),a0 Relocate long pointers
	   lea.l    Mono(pc),a2     to code to handle each
	   move.l   a2,(a0)+	    sample data format
	   lea.l    Stereo(pc),a2
	   move.l   a2,(a0)+
	   lea.l    MonoInt(pc),a2
	   move.l   a2,(a0)+
	   lea.l    StereoInt(pc),a2
	   move.l   a2,(a0)+
	   lea.l    MonoAvg(pc),a2
	   move.l   a2,(a0)+
	   lea.l    StereoAvg(pc),a2
	   move.l   a2,(a0)+
	   lea.l    MonoDec(pc),a2
	   move.l   a2,(a0)+
	   lea.l    StereoDec(pc),a2
	   move.l   a2,(a0)+
	   lea.l    MonoPhase(pc),a2 Interleaved left/right
	   move.l   a2,(a0)+

	   lea.l    io_pointer(pc),a0 Set up device linkage
	   lea.l    io_code(pc),a2  Find absolute address
	   move.l   a2,(a0)+	    Plug it into the table
	   lea.l    open_code(pc),a2
	   move.l   a2,(a0)+	    And so on...
	   lea.l    close_code(pc),a2
	   move.l   a2,(a0)+
	   lea.l    linkage(pc),a0  Link the SOUND device
	   moveq    #32,d0	    MT.LIOD key
	   trap     #1
	   rts			    Return D0 to caller

	   cnop    0,4		    32 bit alignment

* This device is not ROMable because of the following variables

linkage    dc.l    0		    Link to next device
io_pointer dc.l    0		    Pointer to I/O code
	   dc.l    0		    Pointer to OPEN code
	   dc.l    0		    Pointer to CLOSE code

sss_ptr    dc.l    0		    -> 'SSSS' linkage flag
one_byte   dc.w    1
	   dc.w    0		    Space to stuff the odd byte

Routines   dc.l    0		    SOUND1 code address
	   dc.l    0		    SOUND2 code address
	   dc.l    0		    SOUND3 (mono interpolated)
	   dc.l    0		    SOUND4 (stereo interpolated)
	   dc.l    0		    SOUND5 (mono averaged)
	   dc.l    0		    SOUND6 (stereo averaged)
	   dc.l    0		    SOUND7 (mono decimated)
	   dc.l    0		    SOUND8 (stereo decimated)
	   dc.l    0		    SOUND9 (mono interleaved)

* Code to handle TRAP #2 calls: OPEN and CLOSE

close_code move.w  194,a2	    MM.RECHP vector
	   jmp	   (a2) 	    Exit, deallocating channel details

* OPEN is not quite so trivial

open_code  subq.l  #2,a7	    Make space on stack for unit number
	   movea.l a7,a3	    A3 -> space for parameter
	   movea.l a7,a5	    Copy for later (survives MM.ALCHP)
	   move.w  290,a2	    Use IO.NAME vector to decode name
	   jsr	   (a2)

	   bra.s   exit_open	    Jump here if name is faulty
	   bra.s   exit_open	    Jump here if parameter is faulty
	   bra.s   open_ok	    Jump here if name parsed OK

name_spec  dc.w    5		    Length of name
	   dc.b    'SOUND '
	   dc.w    1		    Optional parameter
	   dc.w    -1,1 	    Number suffix, default 1 (mono)

* The open worked; check if we've got a plausible numeric suffix

open_ok    moveq   #28,d1	    24+4 bytes needed
	   move.w  192,a2	    MM.ALCHP vector
	   jsr	   (a2) 	    Allocate memory
	   bne.s   exit_open	    Give up if RAM not found

do_type    moveq   #-7,d0	    Device (suffix) not found?
	   move.w  (a5),d1	    Get channel type suffix
	   beq.s   exit_open	    SOUND0? I think not!
	   cmp.w   #9,d1	    Expect 1 to 9 (for now)
	   bhi.s   exit_open	    That's toooo much (or <=0)

* If we get this far we've got a plausible name and suffix (if any)
* work out the appropriate code to handle output to each stream type
* Pick routine for subsequent data, depending on the numeric suffix

	   lea.l  Routines(pc),a2
	   lsl.w  #2,d1 	    Sift SOUND1..SOUNDn
	   move.l -4(a2,d1.w),sss_type(a0)
	   moveq  #0,d0 	    No error

exit_open  addq.l  #2,a7	    Deallocate stack space
	   tst.l   d0		    Flag error code in D0
	   rts

not_yet    moveq   #-19,d0	    NOT IMPLEMENTED report code
	   rts

* Code to handle TRAP #3 calls: SBYTE, SSTRG, POSRE

io_code    movea.l sss_ptr(pc),a3   Assume SSS is still there
	   move.l  4(a3),a2	    Get base of routines
	   cmp.b   #7,d0	    Call IO.SSTRG?
	   beq.s   send_lots
*
	   cmp.b   #5,d0
	   beq.s   send_one

	   cmp.b   #67,d0	   Call to FS.POSRE?
	   beq.s   position

	   cmp.b   #65,d0	   IOF.FLSH?
	   bne.s   not_yet	   Nothing else is implemented

* Flush device -> kill sound
flush
	   jsr	   16(a2)	   Call SSS_KILL
	   moveq   #0,d0
	   rts

* Trap calls to move the file pointer zero bytes and
* return the current buffer size as if it were a file
* pointer - so FPOS becomes a convenient way to see
* how much space there is in the buffer (it's device
* independent, Jim, but not as we know it...)
position   tst.l   d1		   Are we trying to move?
	   bne.s   bad_param

	   jsr	   20(a2)	   Call SSS_SAMPLE

	   move.l  d0,d1	   Return sample count in queue

* Divide by 20000 to give the length of sound remaining, in seconds
* N.B. Early SSS implementations for Qdos Classic and SMSQ/E return
* the wrong answer to this call, miscalculating by a factor of 2.

happy_go   moveq   #0,d0
	   rts

bad_param  moveq   #-15,d0
	   rts

* SBYTE seems barely worth retaining, just for mono single byte output

send_one   moveq   #1,d2	   Number of bytes
	   lea.l   one_byte(pc),a1
	   move.b  d1,(a1)	   Continue as for SSTRG

* SSTRG is by far the most likely SOUND device call to be performed

send_lots  move.l  sss_type(a0),a5 Sift out according to the stream
	   move.l  a1,a4	   Remember the parameter address
	   move.w  d2,d1	   Remember the total number of bytes
	   beq.s   happy_go	   Zero bytes is trivial

	   jsr	   20(a2)	   Call SSS_SAMPLE
	   swap    d0
	   tst.w   d0
	   beq.s   info1

	   moveq   #0,d1
	   moveq   #-1,d0
	   rts			   exit if more than 3.2 secs in queue

info1
	   jsr	   8(a2)	   Call SSS_SETM
	   exg.l   a1,a4

	   moveq   #0,d0	   Assume it will work out OK
	   jmp	   (a5) 	   Do whatever it takes

* 20 KHz MONO starts in the middle of the loop, checking for space
*
* It might be more efficient to read the input once and write it
* twice, but as most SSS CPUs have data caches and those that do
* not are likely to be code-bandwidth limited, it hardly matters

move1	   move.b  (a1),(a4)+	   Copy left byte to the buffer
	   move.b  (a1)+,(a4)+	   Same again on the right
	   subq.w  #1,d2	   Anything left?
	   bls.s   complete

Mono	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move1

	   bra.s   leave_some

Stereo	   btst    #0,d2	   Odd number of bytes?
	   bne.s   odd_byte	   Stereo data comes in pairs

	   bra.s   nigh2

* The next lines do not assume word alignment, in case the data
* starts at an odd address and we are on a 68000, 68008 or 68010

move2	   move.b  (a1)+,(a4)+	   Copy left to the buffer
	   move.b  (a1)+,(a4)+	   Copy right byte
	   subq.w  #2,d2	   Anything left?
	   bls.s   complete

nigh2	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move2

* Buffer full - timeout, come back for more later, hopefully - this
* part did not seem to work right in the first version of this code

leave_some moveq   #-1,d0	   Leave the rest for later

complete
	   exg.l   a1,a4	   Remember where we got to
	   move.l  4(a3),a2
	   jsr	   12(a2)	   Call SSS_ADDM
	   exg.l   a1,a4	   Remember where we got to

	   sub.w   d2,d1	   Number of bytes outstanding
	   rts

odd_byte   moveq   #-15,d0	   Bad parameter - stereo pair expected
	   rts

* SOUND3 supports half-rate (10 KHz) mono samples by linear interpolation
*	 This gets rather noisy but is far less costly than FIR resampling

MonoInt    move.w  #128,d4	   Start interpolation mid-scale
	   moveq   #0,d5	   Clear high bytes
	   bra.s   nigh3

move3	   move.b  (a1)+,d5	   Get one new data item
	   add.w   d5,d4	   Add last to get new datum
	   lsr.w   #1,d4	   Stay in unsigned byte range
	   move.b  d4,(a4)+	   Copy left byte to the buffer
	   move.b  d4,(a4)+	   Copy right byte to the buffer
	   move.w  d5,d4	   Keep data for next averaging
	   move.b  d4,(a4)+	   Copy left byte to the buffer
	   move.b  d4,(a4)+	   Copy right byte to the buffer
	   subq.w  #1,d2	   Anything left?
	   bls.s   complete

nigh3	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move3
	   bra.s   leave_some

* SOUND4 supports half-rate (10 KHz) stereo samples by linear interpolation

StereoInt  btst    #0,d2	   Odd number of bytes?
	   bne.s   odd_byte

	   move.w  #128,d4	   Start interpolation mid-scale
	   moveq   #0,d5	   Clear high bytes
	   move.w  #128,d6	   Start interpolation mid-scale
	   moveq   #0,d7	   Clear high bytes
	   bra.s   nigh4

move4	   move.b  (a1)+,d5	   Get one new data item
	   add.w   d5,d4	   Add last to get new datum
	   lsr.w   #1,d4	   Stay in unsigned byte range
	   move.b  d4,(a4)+	   Copy left guess to the buffer
	   move.w  d5,d4	   Keep data for next averaging
	   move.b  (a1)+,d6	   Get one new data item
	   add.w   d7,d6	   Add last to get new datum
	   lsr.w   #1,d6	   Stay in unsigned byte range
	   move.b  d6,(a4)+	   Copy right guess to the buffer
	   move.w  d7,d6	   Keep data for next averaging
	   move.b  d4,(a4)+	   Copy left byte to the buffer
	   move.b  d6,(a4)+	   Copy right byte to the buffer
	   subq.w  #2,d2	   Anything left?
	   bls.s   complete

nigh4	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move4
	   bra.s   leave_some

* SOUND5 supports double rate (40 KHz) MONO by averaging sample pairs

MonoAvg    btst    #0,d2	   Odd number of bytes?
	   bne.s   odd_byte

	   moveq   #0,d4	   Clear high bytes
	   moveq   #0,d5
	   bra.s   nigh5

move5	   move.b  (a1)+,d4	   Get one new data item
	   move.b  (a1)+,d5	   Get next datum for averaging
	   add.w   d5,d4	   Sum words
	   lsr.w   #1,d4	   Stay in unsigned byte range
	   move.b  d4,(a4)+	   Copy left byte to the buffer
	   move.b  d4,(a4)+	   Copy right byte to the buffer
	   subq.w  #2,d2	   Anything left?
	   bls.s   complete6

nigh5	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move5
	   bra	   leave_some

* SOUND6 supports double rate (40 KHz) STEREO by averaging sample pairs

StereoAvg  moveq   #3,d4	   Mask for low two bits
	   and.w   d2,d4	   Odd number of bytes or pairs?
	   bne	   odd_byte

	   moveq   #0,d5	   Clear D5 as well as D4 for averages
	   bra.s   nigh6

move6	   move.b  (a1),d4	   Get first left data item
	   move.b  2(a1),d5	   Get next left for averaging
	   add.w   d5,d4	   Sum words
	   lsr.w   #1,d4	   Stay in unsigned byte range
	   move.b  d4,(a4)+	   Copy left byte to the buffer
	   move.b  1(a1),d4	   Get new data item from right
	   addq.l  #3,a1	   Advance to last in this group
	   move.b  (a1)+,d5	   Get last datum for averaging
	   add.w   d5,d4	   Sum words
	   lsr.w   #1,d4	   Stay in unsigned byte range
	   move.b  d4,(a4)+	   Copy right byte to the buffer
	   subq.w  #4,d2	   Anything left?
	   bls.s   complete6

nigh6	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move6
	   bra	   leave_some

complete6  moveq   #0,d2	  Forget any overshoot in decimation
	   bra	   complete	  Keeps loop conditional branches tight

* SOUND7 supports double rate (40 KHz) MONO by skipping alternate samples
* This will yield loads of noisy aliasing distortion if the data is not
* filtered to remove components above 10 KHz, so it no replacement for
* proper resampling - though it is an awful lot less CPU intensive...

move7	   move.b  (a1),(a4)+	   Copy left byte to the buffer
	   move.b  (a1),(a4)+	   Same again on the right
	   addq.l  #2,a1	   Skip the next mono byte
	   subq.w  #2,d2	   Anything left?
	   bls.s   complete6

MonoDec    cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move7
	   bra	   leave_some

* SOUND8 supports double rate (40 KHz) MONO by skipping alternate samples
* See the warnings above about likelyhood of unpleasant aliasing products

StereoDec  btst    #0,d2	   Odd number of bytes?
	   bne	   odd_byte

	   bra.s   nigh8

* The next lines do not assume word alignment, in case the data
* starts at an odd address and we are on a 68000, 68008 or 68010

move8	   move.b  (a1)+,(a4)+	   Copy left to the buffer
	   move.b  (a1)+,(a4)+	   Copy right byte
	   subq.w  #4,d2	   Anything left?
	   bls.s   complete6

	   addq.l  #2,a1	   Skip a stereo pair completely

nigh8	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move8

* SOUND9 supports double rate (40 KHz) MONO by sending alternate
*	 bytes to left and right. This might sound OK if your right
*	 speaker is about half an inch further away than the left!

MonoPhase  btst    #0,d2	   Odd number of bytes?
	   bne	   odd_byte

	   bra.s   nigh9

move9	   move.b  (a1)+,(a4)+	    Copy left byte to the buffer
	   move.b  (a1)+,(a4)+	    Next one on the right
	   subq.w  #2,d2	    Anything left?
	   bls.s   complete6

nigh9	   cmpa.l  a2,a4	   Is the end (of the buffer) nigh?
	   bcs.s   move9

	   bra	   leave_some

	   end
